// Los.h

#ifndef LOS_H_ROGUELIKE_1
#define LOS_H_ROGUELIKE_1

#include "Map.h"

struct Line
{
  Line(Offset newNear=Offset(), Offset newFar=Offset()) : near(newNear), far(newFar) {}
  bool isBelow(Offset const & point)
  {
    return relativeSlope(point) > 0;
  }

  bool isBelowOrContains(Offset const & point)
  {
    return relativeSlope(point) >= 0;
  }

  bool isAbove(Offset const & point)
  {
    return relativeSlope(point) < 0;
  }

  bool isAboveOrContains(Offset const & point)
  {
    return relativeSlope(point) <= 0;
  }

  bool doesContain(Offset const & point)
  {
    return relativeSlope(point) == 0;
  }

  // negative if the line is above the point.
  // positive if the line is below the point.
  // 0 if the line is on the point.
  int relativeSlope(Offset point)
  {
//    if (point.x >= far.x)
//    {
//      return (far.y - near.y)*(point.x - near.x) - (point.y - near.y)*(far.x - near.x);
//    }
//    else
//    {
      return (far.y - near.y)*(far.x - point.x) - (far.y - point.y)*(far.x - near.x);
//    }
  }

  Offset near;
  Offset far;
};

struct Bump
{
  Bump() : parent(NULL) {}
  Offset location;
  Bump * parent;
};

struct Field
{
  Field() : steepBump(NULL), shallowBump(NULL) {}
  Line steep;
  Line shallow;
  Bump * steepBump;
  Bump * shallowBump;
};

// T must implement:
//
// bool operator()(Offset const & source, Offset const & pos);
//
// The result is true if the location is 'blocked' (by the limited
// radius of a lightsource for example).
template <typename T>
class Los
{
public:
  void calculateFovQuadrant(void)
  {
    std::list<Bump> steepBumps;
    std::list<Bump> shallowBumps;
    // activeFields is sorted from shallow-to-steep.
    std::list<Field> activeFields;
    activeFields.push_back(Field());
    activeFields.back().steep.near = Offset(1,0);
    activeFields.back().steep.far = Offset(0,100);
    activeFields.back().shallow.near = Offset(0,1);
    activeFields.back().shallow.far = Offset(100,0);

    std::list<Field>::iterator currentField = activeFields.begin();

    Offset dest(0,0);
    actIsBlocked(dest);
    int distance = 1;
    int i = 0;
    int j = 0;
    // For each square outline
    for (i = 1; i < 100 && ! activeFields.empty(); ++i)
    {
      // Visit the nodes in the outline
      for (j = 0; j <= i; ++j)
      {
        dest.x = i-j;
        dest.y = j;
        visitSquare(dest, currentField, steepBumps, shallowBumps,activeFields);
      }
      currentField = activeFields.begin();
    }
  }

  void visitSquare(Offset dest, std::list<Field>::iterator & currentField,
                   std::list<Bump> & steepBumps,
                   std::list<Bump> & shallowBumps,
                   std::list<Field> & activeFields)
  {
    // The top-left and bottom-right corners of the destination square.
    Offset topLeft(dest.x, dest.y + 1);
    Offset bottomRight(dest.x + 1, dest.y);
    while (currentField != activeFields.end() && currentField->steep.isBelowOrContains(bottomRight))
    {
      // The square is in case 'above'. This means that it is ignored for the currentField. But the steeper fields might need it.
      ++currentField;
    }
    if (currentField == activeFields.end())
    {
      // case ABOVE
      // The square was in case 'above' for all fields. This means that we no longer care about it or any squares in its diagonal rank.
      // TODO: Add notification here for optimization purposes.
      return;
    }

    // Now we check for other cases.
    if (currentField->shallow.isAboveOrContains(topLeft))
    {
      // case BELOW
      // The shallow line is above the extremity of the square, so that square is ignored.
      return;
    }
    // The square is between the lines in some way. This means that we need to visit it and determine whether it is blocked.
    bool isBlocked = actIsBlocked(dest);
    if (!isBlocked)
    {
      // We don't care what case might be left, because this square does not obstruct.
      return;
    }

    if (currentField->shallow.isAbove(bottomRight) && currentField->steep.isBelow(topLeft))
    {
      // case BLOCKING
      // Both lines intersect the square. This current field has ended.
      currentField = activeFields.erase(currentField);
    }
    else if (currentField->shallow.isAbove(bottomRight))
    {
      // case SHALLOW BUMP
      // The square intersects only the shallow line.
      addShallowBump(topLeft, currentField, steepBumps, shallowBumps);
      currentField = checkField(currentField, activeFields);
    }
    else if (currentField->steep.isBelow(topLeft))
    {
      // case STEEP BUMP
      // The square intersects only the steep line.
      addSteepBump(bottomRight, currentField, steepBumps, shallowBumps);
      currentField = checkField(currentField, activeFields);
    }
    else
    {
      // case BETWEEN
      // The square intersects neither line. We need to split into two fields.
      std::list<Field>::iterator steeperField = currentField;
      std::list<Field>::iterator shallowerField = activeFields.insert(currentField, *currentField);
      addSteepBump(bottomRight, shallowerField, steepBumps, shallowBumps);
      checkField(shallowerField, activeFields);
      addShallowBump(topLeft, steeperField, steepBumps, shallowBumps);
      currentField = checkField(steeperField, activeFields);
    }
  }

  std::list<Field>::iterator checkField(std::list<Field>::iterator currentField, std::list<Field> & activeFields)
  {
    std::list<Field>::iterator result = currentField;
    // If the two slopes are colinear, and if they pass through either extremity, remove the field of view.
    if (currentField->shallow.doesContain(currentField->steep.near)
        && currentField->shallow.doesContain(currentField->steep.far)
        && (currentField->shallow.doesContain(Offset(0,1))
            || currentField->shallow.doesContain(Offset(1,0))))
    {
      result = activeFields.erase(currentField);
    }
    return result;
  }

  void addShallowBump(Offset point, std::list<Field>::iterator currentField, std::list<Bump> & steepBumps, std::list<Bump> & shallowBumps)
  {
    // First, the far point of shallow is set to the new point.
    currentField->shallow.far = point;
    // Second, we need to add the new bump to the shallow bump list for future steep bump handling.
    shallowBumps.push_back(Bump());
    shallowBumps.back().location = point;
    shallowBumps.back().parent = currentField->shallowBump;
    currentField->shallowBump = & shallowBumps.back();
    // Now we have too look through the list of steep bumps and see if any of them are below the line.
    // If there are, we need to replace near point too.
    Bump * currentBump = currentField->steepBump;
    while (currentBump != NULL)
    {
      if (currentField->shallow.isAbove(currentBump->location))
      {
        currentField->shallow.near = currentBump->location;
      }
      currentBump = currentBump->parent;
    }
  }

  void addSteepBump(Offset point, std::list<Field>::iterator currentField, std::list<Bump> & steepBumps, std::list<Bump> & shallowBumps)
  {
    currentField->steep.far = point;
    steepBumps.push_back(Bump());
    steepBumps.back().location = point;
    steepBumps.back().parent = currentField->steepBump;
    currentField->steepBump = & steepBumps.back();
    // Now look through the list of shallow bumps and see if any of them are below the line.
    Bump * currentBump = currentField->shallowBump;
    while (currentBump != NULL)
    {
      if (currentField->steep.isBelow(currentBump->location))
      {
        currentField->steep.near = currentBump->location;
      }
      currentBump = currentBump->parent;
    }
  }

  void calculateLos(Offset const & newSource, Offset const & newDest,
                    T & newAction)
  {
    source = newSource;
    action = newAction;
    quadrant = Offset(1,1);
    if (newSource.x > newDest.x)
    {
      quadrant.x = -1;
    }
    if (newSource.y > newDest.y)
    {
      quadrant.y = -1;
    }
    Offset slope(abs(newSource.x - newDest.x), abs(newSource.y - newDest.y));
    traceSlope(slope);
  }

  void calculateFov(Offset const & newSource, T & newAction)
  {
    source = newSource;
    action = newAction;
    static const int quadrantCount = 4;
    static const Offset quadrants[quadrantCount] = {Offset(1, 1),
                                                    Offset(-1, 1),
                                                    Offset(-1, -1),
                                                    Offset(1, -1)};
    int quadrantIndex = 0;
    for (; quadrantIndex < quadrantCount; ++quadrantIndex)
    {
      quadrant = quadrants[quadrantIndex];
      calculateFovQuadrant();
    }
  }
private:
  bool actIsBlocked(Offset const & pos)
  {
    Offset adjustedPos(pos.x*quadrant.x + source.x,
                       pos.y*quadrant.y + source.y);
    return  g::currentMap == NULL || action(source, adjustedPos)
      || (*g::currentMap)[adjustedPos].place.blocked;
  }
private:
  Offset source;
  Offset quadrant;
  T action;
  static const int slopeCount = 256;
};

#endif
