// interface_win_console.cc

#include "lib.h"
#include "interface.h"

#ifdef INTERFACE_WIN_CONSOLE

#include <windows.h>

using namespace std;

namespace
{
  HANDLE inputHandle;
  HANDLE outputHandle;
}

namespace Interface
{
  void init(void)
  {
    // Windows stuff
    inputHandle = GetStdHandle(STD_INPUT_HANDLE);
    outputHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleMode(inputHandle, ENABLE_PROCESSED_INPUT);
    SetConsoleMode(outputHandle, ENABLE_PROCESSED_OUTPUT);
    COORD bufferSize;
    bufferSize.X = 80;
    bufferSize.Y = 25;
    SetConsoleScreenBufferSize(outputHandle, bufferSize);
  }

  void cleanup(void)
  {
  }
}

namespace MapViewer
{
  size_t mapWidth;
  size_t mapHeight;
  vector<CHAR_INFO> mapShow;
  CHAR_INFO tileToChar[TILE_SIMPLE_END];

  void setChar(Tile theTile, char theChar, WORD attributes)
  {
    tileToChar[theTile].Char.AsciiChar = theChar;
    tileToChar[theTile].Attributes = attributes;
  }

  void init(void)
  {
    mapWidth = 61;
    mapHeight = 21;
    mapShow.resize(mapWidth*mapHeight);

    WORD white = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED
      | FOREGROUND_INTENSITY;

    setChar(TILE_UNKNOWN, ' ', BACKGROUND_BLUE | BACKGROUND_GREEN);
//    setChar(TILE_UNKNOWN, '?', FOREGROUND_BLUE | FOREGROUND_GREEN);
    setChar(TILE_PLAYER, '@', white);
    setChar(TILE_WALL, '#', white);
    setChar(TILE_FLOOR, '.', white);
  }

  void cleanup(void)
  {
  }

  void display(void)
  {
    COORD size;
    size.X = static_cast<SHORT>(mapWidth);
    size.Y = static_cast<SHORT>(mapHeight);
    COORD source;
    source.X = 0;
    source.Y = 0;
    SMALL_RECT dest;
    dest.Left = 0;
    dest.Top = 0;
    dest.Right = static_cast<short>(mapWidth)-1;
    dest.Bottom = static_cast<short>(mapHeight)-1;
    WriteConsoleOutput(outputHandle, & mapShow[0], size, source, &dest);
  }

  void hide(void)
  {
  }

  void setTile(int x, int y, Tile theTile)
  {
    assert(x >= 0 && static_cast<size_t>(x) < mapWidth);
    assert(y >= 0 && static_cast<size_t>(y) < mapHeight);
    CHAR_INFO & dest = mapShow[x + y*mapWidth];
    Tile basicTile = static_cast<Tile>(theTile & TILE_SIMPLE_MASK);
    dest = tileToChar[basicTile];
    if (theTile & TILE_LAST_SEEN)
    {
      dest.Attributes &= ~(FOREGROUND_RED | FOREGROUND_INTENSITY);
      dest.Attributes |= FOREGROUND_GREEN | FOREGROUND_BLUE;
    }
//    if (theTile & TILE_FACING)
//    {
//      dest.Attributes &= ~(BACKGROUND_GREEN | BACKGROUND_BLUE
//                           | BACKGROUND_INTENSITY);
//      dest.Attributes |= BACKGROUND_RED;
//    }
  }

  int getWidth(void)
  {
    return static_cast<int>(mapWidth);
  }

  int getHeight(void)
  {
    return static_cast<int>(mapHeight);
  }
}

namespace KeyboardReader
{
  const int tableSize = 256;
  Key vkToKey[tableSize];
  Key asciiToKey[tableSize];

  void init(void)
  {
    for (int i = 0; i < tableSize; ++i)
    {
      vkToKey[i] = KEY_UNKNOWN;
      asciiToKey[i] = KEY_UNKNOWN;
    }
    vkToKey[VK_NUMPAD0] = KEY_NUMPAD_0;
    vkToKey[VK_NUMPAD1] = KEY_NUMPAD_1;
    vkToKey[VK_NUMPAD2] = KEY_NUMPAD_2;
    vkToKey[VK_NUMPAD3] = KEY_NUMPAD_3;
    vkToKey[VK_NUMPAD4] = KEY_NUMPAD_4;
    vkToKey[VK_NUMPAD5] = KEY_NUMPAD_5;
    vkToKey[VK_NUMPAD6] = KEY_NUMPAD_6;
    vkToKey[VK_NUMPAD7] = KEY_NUMPAD_7;
    vkToKey[VK_NUMPAD8] = KEY_NUMPAD_8;
    vkToKey[VK_NUMPAD9] = KEY_NUMPAD_9;
 }

  void cleanup(void)
  {
  }

  Key getNextKey(void)
  {
    Key result = KEY_UNKNOWN;
    while (result == KEY_UNKNOWN)
    {
      INPUT_RECORD in;
      DWORD length = 1;
      DWORD numberRead = 0;
      BOOL success = ReadConsoleInput(inputHandle, &in, length, &numberRead);
      if (success && numberRead == 1 && in.EventType == KEY_EVENT
          && in.Event.KeyEvent.bKeyDown)
      {
        WORD vKeycode = in.Event.KeyEvent.wVirtualKeyCode;
        CHAR ascii = in.Event.KeyEvent.uChar.AsciiChar;
        DWORD control = in.Event.KeyEvent.dwControlKeyState;
        result = vkToKey[vKeycode];
        if (result == KEY_UNKNOWN)
        {
          result = asciiToKey[ascii];
        }
        if (result != KEY_UNKNOWN)
        {
          if ((control & CAPSLOCK_ON) || (control & SHIFT_PRESSED))
          {
            result = static_cast<Key>(result | KEY_SHIFT);
          }
          if ((control & LEFT_CTRL_PRESSED) || (control & RIGHT_CTRL_PRESSED))
          {
            result = static_cast<Key>(result | KEY_CONTROL);
          }
          if ((control & LEFT_ALT_PRESSED) || (control & RIGHT_ALT_PRESSED))
          {
            result = static_cast<Key>(result | KEY_ALT);
          }
        }
      }
    }
    return result;
  }
}

#endif
